% Performs the numerical tests from Section 4.2, 4.3 and 4.4 in
%
%   L. Kaemmerer, D. Potts and F. Taubert
%   The uniform sparse FFT with application to PDEs with random coefficients
%   Submitted, 2021
%
% Use the parameter 'example_name' to choose the particular example, you
% want to consider. In particular, the possible choices are:
%
%   'periodic_1' - the example from Sec. 4.2 with c = 0.4 and \mu = 1.2
%   'periodic_2' - the example from Sec. 4.2 with c = 1.5 and \mu = 3.6
%   'affine'     - the example from Sec. 4.3
%   'lognormal'  - the example from Sec. 4.4
%
% Note: The Partial Differential Equation Toolbox of MATLAB is necessary!

%% PDE initialization and the usFFT
% Note: If the variables 'freq' and 'val' still exist in the workspace, 
% this part is skipped by default to avoid an accidentally calling of the
% usFFT and the deletion the current results.
% If you want to force the call of the usFFT (and change the PDE and usFFT
% parameters), clear the variables 'freq' and/or 'val'.

% Set various parameters.

if (~exist('freq','var') || ~exist('val','var'))
    
    example_name = 'affine'; % specifies the example, see explanation above
    N = 32; % Extension of the search space Gamma = [-N,N]^d
    s = 100; % Sparsity
    
    r = 5; % number of detection iterations, cf. usfft_algorithm.m
    theta = 1e-12; % threshold, cf. usfft_algorithm.m
    prob_delta = 0.1; % probability parameter, cf. usfft_algorithm.m
    hmax = 0.075; % FEM parameter (target maximum mesh edge length)

% Load PDE model and perform the usFFT.

    switch example_name   
        case 'periodic_1'
            model = 'periodic';
            d = 10; 
            a0 = 1;
            c = 0.4;
            mu = 1.2;
            [nodes,freq,val,freq_cand_sizes,sample_sizes,run_times]=usfft_algorithm(@(y) solve_pde_per(y,a0,c,mu,hmax),d,N,'rlfftriabs',s,s,theta,theta,r,r,prob_delta);

        case 'periodic_2'
            model = 'periodic';
            d = 10; 
            a0 = 1;
            c = 1.5;
            mu = 3.6;       
            [nodes,freq,val,freq_cand_sizes,sample_sizes,run_times]=usfft_algorithm(@(y) solve_pde_per(y,a0,c,mu,hmax),d,N,'rlfftriabs',s,s,theta,theta,r,r,prob_delta);

        case 'affine'
            model = 'affine';
            d = 20; 
            a0 = 1;
            alpha0 = 0.9/zeta(2);
            mu = 2;
            alpha = -1;
            beta = 1;
            [nodes,freq,val,freq_cand_sizes,sample_sizes,run_times]=usfft_algorithm(@(y) solve_pde_aff(y,a0,alpha0,mu,hmax,'tent'),d,N,'rlfftriabs',s,s,theta,theta,r,r,prob_delta);

        case 'lognormal'
            model = 'lognormal';
            d = 10; 
            a0 = 0;
            temp_1 = max(2*N+1, floor(7*s));
            temp_2 = simplenextprime(temp_1);
            Delta = 1/(4*temp_2);
            [nodes,freq,val,freq_cand_sizes,sample_sizes,run_times]=usfft_algorithm(@(y) solve_pde_log(y,a0,hmax,'tent_err',Delta),d,N,'rlfftriabs',s,s,theta,theta,r,r,prob_delta);

        otherwise
            error('unknown example: %s', example_name);
    end
    
end

%% l_p error estimation
% Set test parameters

n_test = 10^4; % number of test samples for the error estimation
p = 2; % specifies l_p norm to be used for the error estimation

% Generate samples of the "exact" solution and the approximation.

G = size(nodes,2);
u = zeros(n_test,G);

switch example_name
    case {'periodic_1','periodic_2'}  
        y = rand(n_test,d)-0.5;
        [~,u_exact] = solve_pde_per(y,a0,c,mu,hmax); 
        parfor j=1:n_test            
            for i = 1:G
                u(j,i) = u_val(y(j,:),freq,val(:,i),model);
            end
        end
            
    case 'affine'
        y = 2*rand(n_test,d)-1;
        [~,u_exact] = solve_pde_aff(y,a0,alpha0,mu,hmax,'notrans');  
        parfor j=1:n_test            
            for i = 1:G
                u(j,i) = u_val(y(j,:),freq,val(:,i),model,alpha,beta);
            end
        end
        
    case 'lognormal'
        y = randn(n_test,d);
        [~,u_exact] = solve_pde_log(y,a0,hmax,'notrans',Delta);
        parfor j=1:n_test            
            for i = 1:G
                u(j,i) = u_val(y(j,:),freq,val(:,i),model,nan,nan,Delta);
            end
        end
        
    otherwise
        error('unknown example: %s', example_name);
end

% Set values at boundary nodes to exactly zero.

outer_ind = find(any(nodes==1 | nodes==0,1));
%outer_nodes = nodes(:,outer_ind);
u_exact(:,outer_ind) = zeros(n_test,length(outer_ind));

% Compute the error in the given l_p norm.

u_err = abs(u-u_exact);
err = zeros(G,1);

if p==inf
    err = max(u_err,[],1);
else
    for j = 1:G
        err(j) = (n_test)^(-1/p)*norm(u_err(:,j),p);
    end
end

% Plot the error.

figure();
F = scatteredInterpolant(nodes(1,:)',nodes(2,:)',err);
[X,Y]=meshgrid(linspace(0,1,100));
myplot = pcolor(X,Y,F(X,Y));
set(myplot, 'EdgeColor', 'none');
myplot.FaceColor = 'interp';
hold on;
scatter(nodes(1,:),nodes(2,:),1,'filled','MarkerEdgeColor',[0,0,0],'MarkerFaceColor',[0,0,0]);
colormap jet
colorbar
xticks([0, 0.5, 1]);
yticks([0, 0.5, 1]);
axis square;
ax = gca;
ax.FontSize = 16;

%% EW estimation
% Compute expectation value, load the corresponding Monte Carlo 
% approximation (generated with the script montecarlo_ev.m, an 
% approximation using 10^6 samples is already given for each example)
% and compute the error.

switch example_name
    case 'periodic_1'
        expval = real(exp_val(freq,val,model))';
        mc_ev = readmatrix('montecarlo_results/periodic1.csv');
        
    case 'periodic_2'
        expval = real(exp_val(freq,val,model))'; 
        mc_ev = readmatrix('montecarlo_results/periodic2.csv');
        
    case 'affine'
        expval = real(exp_val(freq,val,model))';
        mc_ev = readmatrix('montecarlo_results/affine.csv');
        
    case 'lognormal'
        expval = real(exp_val(freq,val,model,Delta))';
        mc_ev = readmatrix('montecarlo_results/lognormal.csv');
        
    otherwise
        error('unknown example: %s', example_name);
end

ev_err = abs(expval-mc_ev);

% Plot the error.

figure()
F = scatteredInterpolant(nodes(1,:)',nodes(2,:)',ev_err);
[X,Y]=meshgrid(linspace(0,1,100));
myplot = pcolor(X,Y,F(X,Y));
set(myplot, 'EdgeColor', 'none');
myplot.FaceColor = 'interp';
hold on;
scatter(nodes(1,:),nodes(2,:),1,'filled','MarkerEdgeColor',[0,0,0],'MarkerFaceColor',[0,0,0]);
colormap jet
colorbar
xticks([0, 0.5, 1]);
yticks([0, 0.5, 1]);
axis square;
ax = gca;
ax.FontSize = 16;

%% ANOVA decomposition
% Count the amount of coefficients with exactly l non-zero components.

h = sum(freq./freq,2,'omitnan');
ind = 1:1:max(h);
amounts = histcounts(h);

% Compute variances and global sensitivity indices.

D = zeros(max(h)+1,G);
total_var = zeros(G,1);
gsi = zeros(max(h),G);

for j = 1:G
    D(:,j) = accumarray(h+1, abs(val(:,j)).^2);
    total_var(j) = sum(D(2:end,j));
    gsi(:,j) = D(2:end,j)./total_var(j);
end

max_gsi = max(gsi,[],2);
min_gsi = min(gsi,[],2);

% Plot the results.

figure()
set(gcf, 'Position',  [100, 200, 1600, 800])
yyaxis left
b = bar(ind-0.2,amounts(2:length(amounts)),0.35);
xlabel('number of non-zero frequency components')
ylabel('number of detected frequencies')
xlim([0.5,max(h)+0.5])
xticks(ind)
grid on
xtips1 = b(1).XEndPoints;
ytips1 = b(1).YEndPoints;
labels1 = string(b(1).YData);
text(xtips1,ytips1,labels1,'HorizontalAlignment','center','VerticalAlignment','bottom','FontSize',12)

yyaxis right
b1 = bar(ind+0.2,max_gsi,0.35);
ylabel('global sensitivity indices')
grid on
set(gca,'yScale','log')
xtips2 = b1(1).XEndPoints;
ytips2 = b1(1).YEndPoints;
labels2 = string(b1(1).YData);
text(xtips2,ytips2,labels2,'HorizontalAlignment','center','VerticalAlignment','bottom','FontSize',12)
hold on

p1 = plot([1+0.025, 1+0.375],[mean(gsi(1,:),2,'omitnan'),mean(gsi(1,:),2,'omitnan')],'k--','DisplayName','mean');
hold on
p2 = plot([1+0.025, 1+0.375],[median(gsi(1,:),2,'omitnan'),median(gsi(1,:),2,'omitnan')],'k:','DisplayName','median');
hold on

for j = 2:max(h)
   plot([j+0.025, j+0.375],[mean(gsi(j,:),2,'omitnan'),mean(gsi(j,:),2,'omitnan')],'k--');
   hold on
   plot([j+0.025, j+0.375],[median(gsi(j,:),2,'omitnan'),median(gsi(j,:),2,'omitnan')],'k:');
   hold on
end

b2 = bar(ind+0.2,min_gsi,0.35);
set(b2(1),'facecolor',[0.9290, 0.6940, 0.1250])
grid on
set(gca,'yScale','log')
xtips2 = b2(1).XEndPoints;
ytips2 = b2(1).YEndPoints;
labels2 = string(b2(1).YData);
text(xtips2,ytips2,labels2,'HorizontalAlignment','center','VerticalAlignment','top','FontSize',12)

legend([p1,p2])
ax = gca;
ax.FontSize = 16;