function[f]=exp_val(freq,val,model_name,delta)
% EXP_VAL Computes the expectation value of the solution, cf. Section 4.1 in 
% 
%   L. Kaemmerer, D. Potts and F. Taubert
%   The uniform sparse FFT with application to PDEs with random coefficients
%   Submitted, 2021
%
%   Input parameters:
%      freq - frequency index set
%      val - the corresponding Fourier coefficients
%      model_name - decodes if a periodization was used
%                   'periodic' - no periodization used
%                   'affine' - tent transform used, cf. Sec. 3.2.1
%                   'lognormal' - lognormal transform used, cf. Sec. 3.2.2
%      delta - (optional) lognormal transformation parameter,
%              necessary if model_name = 'lognormal'         
%
% Output parameters: f - expectation value

if nargin < 4 || isempty(delta)
  delta = nan;
end

f = zeros([1,size(val,2)]);
[s,d] = size(freq);

switch model_name
    case 'periodic'
        f = val(sum(freq./freq,2,'omitnan')==0,:);
    case 'affine'
        for j1=1:s
            c=1;
            for j2=1:d
                if(mod(freq(j1,j2),2)==1)
                    fak(j2)=2*1i/(pi*freq(j1,j2));
                elseif(freq(j1,j2)==0)
                    fak(j2)=1;
                else
                    fak(j2)=0;
                end
                c=c*fak(j2); 
            end
            f=f+c*val(j1,:);
        end
    case 'lognormal'  
        if isnan(delta)
            error('delta is needed in the lognormal case');
        end
        for j1=1:s
            c=1;
            for j2=1:d
                if(mod(freq(j1,j2),2)==1)
                    fak(j2)=1i*2/(pi*freq(j1,j2))*exp(2*pi*1i*freq(j1,j2)*delta);
                elseif(freq(j1,j2)==0)
                    fak(j2)=1;
                else
                    fak(j2)=0;
                end
                c=c*fak(j2); 
            end
            f=f+c*val(j1,:);
        end
    otherwise
        error('unsupported model: %s', model_name);
end
end