% Computes Monte Carlo approximations of the expectation value of the
% solution of the examples from Section 4.2, 4.3 and 4.4 in
%
%   L. Kaemmerer, D. Potts and F. Taubert
%   The uniform sparse FFT with application to PDEs with random coefficients
%   Submitted, 2021
%
% Note: The Partial Differential Equation Toolbox of MATLAB is necessary!

% Initialization.

n_test = 10^6; % number of samples for the Monte Carlo approximation

hmax = 0.075; % FEM parameter (target maximum mesh edge length)

example_name = 'lognormal'; %

switch example_name
    case 'periodic_1'
        model = 'periodic';
        d = 10; 
        a0 = 1;
        c = 0.4;
        mu = 1.2; 
        
    case 'periodic_2'
        model = 'periodic';
        d = 10; 
        a0 = 1;
        c = 1.5;
        mu = 3.6;  

    case 'affine'
        model = 'affine';
        d = 20; 
        a0 = 1;
        alpha0 = 0.9/zeta(2);
        mu = 2;
        
    case 'lognormal'
        model = 'lognormal';
        d = 10; 
        a0 = 0;
        
    otherwise
        error('unknown example: %s', example_name);
end

% Sampling and computation of the approximation.

switch example_name
    case {'periodic_1','periodic_2'}  
        y = rand(n_test,d)-0.5;
        [nodes,u_exact] = solve_pde_per(y,a0,c,mu,hmax); 
            
    case 'affine'
        y = 2*rand(n_test,d)-1;
        [nodes,u_exact] = solve_pde_aff(y,a0,alpha0,mu,hmax,'notrans');  
        
    case 'lognormal'
        y = randn(n_test,d);
        [nodes,u_exact] = solve_pde_log(y,a0,hmax,'notrans');
        
    otherwise
        error('unknown example: %s', example_name);
end

outer_ind = find(any(nodes==1 | nodes==0,1));
G = size(nodes,2);
mc_ev = zeros(G,1);

parfor g = 1:G
    if ismember(g,outer_ind)
        mc_ev(g) = 0;
    else
        mc_ev(g) = mean(u_exact(:,g));
    end
end

% Plot the approximation.

figure()
F = scatteredInterpolant(nodes(1,:)',nodes(2,:)',mc_ev);
[X,Y]=meshgrid(linspace(0,1,100));
myplot = pcolor(X,Y,F(X,Y));
set(myplot, 'EdgeColor', 'none');
myplot.FaceColor = 'interp';
hold on;
scatter(nodes(1,:),nodes(2,:),1,'filled','MarkerEdgeColor',[0,0,0],'MarkerFaceColor',[0,0,0]);
colormap jet
colorbar
xticks([0, 0.5, 1]);
yticks([0, 0.5, 1]);
axis square;
ax = gca;
ax.FontSize = 16;

% Save the approximation.

% writematrix(mc_ev,'montecarlo_results/lognormal.csv')