function [ Ms,zs ] = multi_lattice_construct_mr1l(I,algno,arg)
% MULTI_LATTICE_CONSTRUCT_MR1L_2 builds multiple reconstructing rank-1 lattice configuration for index set I.
%
%   algno==2 corresponds to Algorithm 4 from
%      L. Kaemmerer
%      Constructing spatial discretizations for sparse multivariate trigonometric polynomials that allow for a fast discrete Fourier transform.
%      Appl. Comput. Harmon. Anal., 2017, doi:10.1016/j.acha.2017.11.008.
%
%   Input parameters:
%      I - index set, array of size num_frequencies x spatial_dimensions
%      algno - use 2
%      arg - struct with field 'c' and 'delta';
%            if unsure, use 'arg.c = 2' and 'arg.delta = 0.1'
%
%   Output parameters:
%      Ms - lattice sizes
%      zs - generating vectors
%
% Copyright (c) 2014-2020 Lutz Kaemmerer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

I=double(I);
if check_arg(arg,algno)
    switch algno
        case 2
            [Ms,zs]=construct_mr1l_2(I,arg.c,arg.delta);
      otherwise
        error('Algorithm not supported');
    end
end
end

function a=check_arg(arg, algno)

    a=1;
    switch algno
        case 2
            if(~isfield(arg,'c') || arg.c<=1)
               error('You need to set arg.c>1');
            end
            if(~isfield(arg,'delta') || arg.delta<=0 || arg.delta>1)
               error('You need to set a real value arg.delta \in (0,1)');
            end
        otherwise
            error('The Algorithm no. %d is unknown', algno);
    end
end

function [Ms,zs]=construct_mr1l_2(I,c,delta)

    T=size(I,1);
    N_I=max(max(I)-min(I));
%    c=max([c,N_I/(T-1)]);
    s=ceil((c/(c-1))^2*(log(T)-log(delta))/2);
%    lambda=max(c*(T-1),4*s*log(s));
    lambda=c*(T-1);
    % upper bound on the needed primes
    p0=max([lambda,3275,N_I]);
    pn_ub=(1+1/(2*(log(p0))^2))^s*p0;
    prime_cands=primes(pn_ub);
    prime_cands=prime_cands(prime_cands>lambda);
    PIlambda_n=sort(get_PIlambda_n(I,prime_cands,s));
    
    ell=0;
    M=[];
    z=[];
    tilde_I=zeros(0,size(I,2));
    while size(I,1)>size(tilde_I,1)
       ell=ell+1;
       Ms(ell)=PIlambda_n(ell);
       zs(ell,:)=randi(Ms(ell),1,size(I,2))-1;
       tmp=mod(I*zs(ell,:)',Ms(ell));
       [tmpsort,ind1]=sort(tmp);
       tmpsortdiff=tmpsort(2:end)-tmpsort(1:end-1);
       ind2=[];
       if(numel(tmpsortdiff)==0)||(tmpsortdiff(1)~=0)
           ind2=[ind2;1];
       end
       if(numel(tmpsortdiff)~=0)&&(tmpsortdiff(end)~=0)
           ind2=[ind2;length(tmpsort)];
       end
       ind2tmp=find(tmpsortdiff(1:end-1).*tmpsortdiff(2:end)~=0);
       ind2=[ind2;ind2tmp+1];
       stI=size(tilde_I,1);
       tilde_I=unique([tilde_I;I(ind1(ind2),:)],'rows');
       if(stI==size(tilde_I,1))
           Ms(ell)=[];
           zs(ell,:)=[];
           ell=ell-1;
       end
       if(ell==s)
            if(size(tilde_I,1)~=size(I,1))
                warning('Algorithm 2 did not determine an reco mr1l.'); 
                pause;
            end
            break;
       end
    end
end

function ps=get_PIlambda_n(I,prime_cands,n)

    prime_cands=sort(prime_cands);
    ps=zeros(1,n);
    zaehler=1;
    N_I=max(max(I)-min(I));
    for  j=1:length(prime_cands)
        if(prime_cands(j)>N_I || size(unique(mod(I,prime_cands(j)),'rows'),1)==size(I,1))
            ps(zaehler)=prime_cands(j);
            zaehler=zaehler+1;
        end
        if(zaehler==n+1)
            break;
        end
    end
end