function [freq,grid_size] = reco_cardI_basic1_delta(fct_handle, d, N, dims, threshold)
% RECO_CARDI_BASIC1_DELTA Utility function for the function usfft_algorithm.
%   This function realizes step 1 and 2a of Algorithm 3.1.
%   The one-dimensionally projected Fourier coefficients, which are
%   projected to the component specificied by the parameter dims, are computed
%   and the projected Fourier coefficients larger than the absolute threshold
%   are returned with the corresponding frequencies.
%
%   Input parameters:
%      fct_handle - function handle (black box) which accepts sampling
%                   nodes as an array of size [num_nodes,d]
%      d - dimensionality
%      N - parameter for search space Gamma = [-N,...,N]^d of possible
%          frequencies
%      dims - the component to project to
%      threshold - (optional) relative threshold theta used for truncation
%
%    Ouput parameters:
%       freq - frequencies belonging to the projected Fourier coefficients
%              larger than the absolute threshold
%       val - projected Fourier coefficients larger than the absolute threshold
%
% Copyright (c) 2021 Fabian Taubert
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

if d < 1
  error('d must be at least 1');
end

if ((d > 3) && (nargin < 4))
  error('if d is larger than 3, dims have to be specified');
end

if nargin > 3
  dims = unique(dims);
  if min(dims) < 1
    error('values in dims must be at least 1');
  end
  if max(dims) > d
    error('values in dims must not be greater than d');
  end
else
  if d == 1
    dims = 1;
  elseif d == 2
    dims = [1 2];
  else
    dims = [1 2 3];
  end
end

if nargin < 5
  threshold = 1e-12;
end

if length(dims) == 1
  freq_cand = (-N:N)';
  x = freq_cand / (2*N+1);
else
  error('length(dims) > 1 not supported');
end

newx = repmat(rand(1,d),size(x,1),1);
for t = 1:length(dims)
  newx(:,dims(t)) = x(:,t);
end
x = mod(newx,1);
clear newx;
[nodes,samples_nodal_full] = fct_handle(x);
inner_ind = find(all([nodes(:,:) == 1 ; nodes(:,:) == 0]==0));
inner_nodes = nodes(:,inner_ind);
samples_grid = samples_nodal_full(:,inner_ind);
grid_size = size(inner_nodes,2);
clear x;

ind = [];

for g = 1:grid_size
    samples = samples_grid(:,g);
    val_all = fftshift(fft(ifftshift(samples))) / (2*N+1);
    clear samples;
    
    absolute_threshold = threshold;
    ind_local = find(abs(val_all) >= absolute_threshold);
    ind = union(ind,ind_local);
    clear ind_local
end

freq = freq_cand(ind,:);
end