function [I_tilde, p_hat_tilde, num_updated, num_pfp_removed] = sfft_random_r1l_alg2l4(Gamma, z_matrix, M_array, samples_cell_array, tau)
% SFFT_RANDOM_R1L_ALG2 Reconstruction of sparse high-dimensional trigonometric polynomials
% based on samples along random rank-1 lattices (non dimension-incremental
% basic algorithm), slightly modified version of Algorithm 2, cf. Remark 9 
% in 
%
%   L. Kaemmerer, F. Krahmer and T. Volkmer
%   A sample efficient sparse FFT for arbitrary frequency candidate sets in
%   high dimensions
%   Submitted, 2020
%
%   Input parameters:
%      Gamma - search space of frequencies, array of size [num_freq,d]
%      z_matrix - generating vectors of random rank-1 lattice as
%                 array of size [L,d]
%      M_array - lattice sizes of random rank-1 lattices
%                as vector of length L
%      samples_cell_array - samples of function under consideration
%                           as cell array with L entries,
%                           i-th entry is vector of length M_array(i)
%      tau - threshold parameter for Fourier coefficient detection,
%            use 1e-12 if unsure
%
%   Output parameters:
%      Gamma_I - detected frequencies as array of size [sparsity_s,d]
%      p_hat_tilde - corresponding Fourier coefficients
%
% Copyright (c) 2020 Toni Volkmer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

[num_freq_cand, d ] = size(Gamma);
L = numel(M_array);

if floor((L+1)/2) ~= (L+1)/2
  error('Number L of rank-1 lattices must be odd natural number');
end

if d ~= size(z_matrix,2)
  error('Gamma or z_matrix has wrong size');
end

if L ~= size(z_matrix,1)
  error('z_matrix or M_array has wrong size');
end

if numel(samples_cell_array) ~= L
  error('M_array or samples_cell_array has wrong size');
end

g_hat_multi = nan(num_freq_cand, L);

for ell=1:L
  M_ell = M_array(ell);
  samples_ell = samples_cell_array{ell}(:);
  if M_ell ~= numel(samples_ell)
    error('M_array(%d) or samples_cell_array{%d} has wrong size', ell, ell);
  end

  g_hat_ell = 1/M_ell * fft(samples_ell);
  
  g_hat_multi(:, ell) = g_hat_ell(mod(Gamma*z_matrix(ell,:).',M_ell) + 1);
end

detection_matrix = (abs(g_hat_multi) >= tau);
detection_array = sum(detection_matrix, 2);

ind_detected = find(detection_array >= L/2);

I_tilde = Gamma(ind_detected, :);

g_hat_detected_sorted_real = sort(real(g_hat_multi(ind_detected, :)), 2);
g_hat_detected_sorted_imag = sort(imag(g_hat_multi(ind_detected, :)), 2);
p_hat_tilde = g_hat_detected_sorted_real(:,(L+1)/2) + 1i * g_hat_detected_sorted_imag(:,(L+1)/2);

p_hat_tilde_update_sum = zeros(length(ind_detected),1);
p_hat_tilde_update_cnt = zeros(length(ind_detected),1);

for ell=1:L
  ind_reco_L = get_r1l_reconstructable_freq_indices(I_tilde,z_matrix(ell,:),M_array(ell));
  p_hat_tilde_update_sum(ind_reco_L) = p_hat_tilde_update_sum(ind_reco_L) + g_hat_multi(ind_detected(ind_reco_L), ell);
  p_hat_tilde_update_cnt(ind_reco_L) = p_hat_tilde_update_cnt(ind_reco_L) + 1;
end

ind_updated = find(p_hat_tilde_update_cnt > 0);
num_updated = length(ind_updated);
p_hat_tilde(ind_updated) = p_hat_tilde_update_sum(ind_updated) ./ p_hat_tilde_update_cnt(ind_updated);

ind_to_remove = find(abs(p_hat_tilde(ind_updated)) < tau);
I_tilde(ind_updated(ind_to_remove),:) = [];
p_hat_tilde(ind_updated(ind_to_remove)) = [];
num_pfp_removed = length(ind_to_remove);

end