function [nodes,values] = solve_pde_aff(y,a0,alpha0,mu,hmax,transform)
% SOLVE_PDE_AFF method for solving the PDE from Section 4.3
% The method uses the Partial Differential Equation Toolbox of MATLAB.
%
%   Input parameters:
%      y - sampling locations, size [#amount,d]
%      a0,alpha0,mu - PDE parameters
%      hmax - FEM parameter (regulating the grid size)
%      transform - specifies if a periodization should be applied
%                  'notrans' - no periodization
%                  'tent' - applies the tent transform to all values in y
%                           before solving the PDE
%
%   Output parameters:
%      nodes - nodes of the FEM
%      values - function values of the solution, size [#amount,#nodes]
%
% Copyright (c) 2021 Fabian Taubert

[n,d] = size(y);
switch transform
    case 'notrans'
    case 'tent'
        y = 1-abs(2-4*y);
    otherwise
        error('unsupported transform: %s', transform);
end

model = createpde();
R = [3,4,0,1,1,0,0,0,1,1]';
gm = R;
sf = 'R';
ns = char('R');
g = decsg(gm,sf,ns);
geometryFromEdges(model,g);
applyBoundaryCondition(model,'dirichlet','Edge',1:model.Geometry.NumEdges,'u',0);
specifyCoefficients(model,'m',0,'d',0,'c',@(location, state) dif_coef_pde_aff(location,state,y(1,:),d,a0,alpha0,mu),'a',0,'f',1);
generateMesh(model,'Hmax',hmax);
results = solvepde(model);
nodes = results.Mesh.Nodes;
grid_size = size(nodes,2);
values = zeros(n,grid_size);
values(1,:) = results.NodalSolution;

parfor j = 2:n
    model = createpde();
    geometryFromEdges(model,g);
    applyBoundaryCondition(model,'dirichlet','Edge',1:model.Geometry.NumEdges,'u',0);
    specifyCoefficients(model,'m',0,'d',0,'c',@(location, state) dif_coef_pde_aff(location,state,y(j,:),d,a0,alpha0,mu),'a',0,'f',1);
    generateMesh(model,'Hmax',hmax);
    results = solvepde(model);
    values(j,:) = results.NodalSolution;
end
end

function vlt=dif_coef_pde_aff(location,state,y,d,a0,alpha0,mu)
vlt = a0;
 for j=1:d
  vlt=vlt+y(j).*(alpha0/(j^mu)).*cos(2*pi*beta_1(j)*location.x).*cos(2*pi*beta_2(j)*location.y);
 end
end

function res=beta_1(m)
    res = m - k_value(m) * (k_value(m)+1) * 0.5;
end

function res=beta_2(m)
    res = k_value(m) - beta_1(m);
end

function res=k_value(m)
    res = floor(-1/2 + sqrt(1/4 + 2*m));
end