function [nodes,values] = solve_pde_log(y,a0,hmax,transform,delta)
% SOLVE_PDE_LOG method for solving the PDE from Section 4.4
% The method uses the Partial Differential Equation Toolbox of MATLAB.
%
%   Input parameters:
%      y - sampling locations, size [#amount,d]
%      a0 - PDE parameter
%      hmax - FEM parameter (regulating the grid size)
%      transform - specifies if a periodization should be applied
%                  'notrans' - no periodization
%                  'tent_err' - applies the lognormal transformation from 
%                               Section 3.2.2 to all values in y before 
%                               solving the PDE
%
%   Output parameters:
%      nodes - nodes of the FEM
%      values - function values of the solution, size [#amount,#nodes]
%
% Copyright (c) 2021 Fabian Taubert

[n,d] = size(y);
switch transform
    case 'notrans'
    case 'tent_err'
        y = transfo(y,delta);
    otherwise
        error('unsupported transform: %s', transform);
end

model = createpde();
R = [3,4,0,1,1,0,0,0,1,1]';
gm = R;
sf = 'R';
ns = char('R');
g = decsg(gm,sf,ns);
geometryFromEdges(model,g);
applyBoundaryCondition(model,'dirichlet','Edge',1:model.Geometry.NumEdges,'u',0);
specifyCoefficients(model,'m',0,'d',0,'c',@(location, state) dif_coef_pde_log(location,state,y(1,:),d,a0),'a',0,'f',@(location, state) sin(1.3*pi*location.x+3.4*pi*location.y).*cos(4.3*pi*location.x-3.1*pi*location.y));
generateMesh(model,'Hmax',hmax);
results = solvepde(model);
nodes = results.Mesh.Nodes;
grid_size = size(nodes,2);
values = zeros(n,grid_size);
values(1,:) = results.NodalSolution;

parfor j = 2:n
    model = createpde();
    geometryFromEdges(model,g);
    applyBoundaryCondition(model,'dirichlet','Edge',1:model.Geometry.NumEdges,'u',0);
    specifyCoefficients(model,'m',0,'d',0,'c',@(location, state) dif_coef_pde_log(location,state,y(j,:),d,a0),'a',0,'f',@(location, state) sin(1.3*pi*location.x+3.4*pi*location.y).*cos(4.3*pi*location.x-3.1*pi*location.y));
    generateMesh(model,'Hmax',hmax);
    results = solvepde(model);
    values(j,:) = results.NodalSolution;
end
end

function z = transfo(x,delta)
    y = x;
    for i=1:size(x,1)
       for j=1:size(x,2) 
        xi = x(i,j);
        if xi < delta
            y(i,j) = -1/2 - 2*(xi-delta);
        elseif xi < 1/2 + delta
            y(i,j) = -1/2 + 2*(xi-delta);
        else
            y(i,j) = +3/2 - 2*(xi-delta);
        end
       end
    end   
    z = sqrt(2) * erfinv(2*y);
end

function vlt=dif_coef_pde_log(location,state,y,d,a0)
 vlt = a0;
 for j=1:d
  vlt=vlt+y(j)./j.*sin(2*pi*j*location.x).*cos(2*pi*(11-j)*location.y);
 end
 vlt = exp(vlt);
end