function u = u_val(y,freq,val,model_name,alpha,beta,delta)
% U_VAL Evaluates the trigonometric polynomial at y.
% The method considers eventually used transforms.
%
%   Input parameters:
%      y - value to evaluate the trig. polynomial for
%      freq - frequency index set
%      val - the corresponding Fourier coefficients
%      model_name - decodes if a periodization was used
%                   'periodic' - no periodization used
%                   'affine' - tent transform used, cf. Sec. 3.2.1
%                   'lognormal' - lognormal transform used, cf. Sec. 3.2.2
%      alpha, beta - (optional) affine transformation parameters,
%                    necessary if model_name = 'affine'
%      delta - (optional) lognormal transformation parameter,
%                    necessary if model_name = 'lognormal'
%
%   Output parameter:
%      u - function value of the trig. polynomial at y
%
% Copyright (c) 2021 Fabian Taubert
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

if nargin < 5 || isempty(alpha)
  alpha = nan;
end

if nargin < 6 || isempty(beta)
  beta = nan;
end

if nargin < 7 || isempty(delta)
  delta = nan;
end

d = length(y);
u = 0;

switch model_name
    case 'periodic'
        for j=1:size(val)
            s=val(j);
            for k=1:d
                s=s.*exp(2*pi*1i*freq(j,k).*y(k));
            end  
            u=u+s;
        end
    case 'affine'
        if (isnan(alpha)) || (isnan(beta))
            error('alpha and beta are needed in the affine case');
        end
        for j=1:size(val)
            s=val(j);
            for k=1:d
                s=s.*exp(pi*1i/(beta-alpha)*freq(j,k).*(y(k)-alpha));
            end  
        u=u+s;
        end
    case 'lognormal'
        if isnan(delta)
            error('delta is needed in the lognormal case');
        end
        for j=1:size(val)
            s=val(j);
            for k=1:d
                s=s.*exp(2*pi*1i*freq(j,k).*(erf(y(k)/sqrt(2))/4+delta+1/4));
            end  
            u=u+s;
        end
    otherwise
        error('unsupported model: %s', model_name);
end

u = real(u);
end