function [nodes, freq, val, freq_cand_sizes, sample_sizes, run_times] = usfft_algorithm(fct_handle, d, N, algorithm_name, sparsity_s, sparsity_s_local, threshold_theta, threshold_theta_basic, niter_r, niter_r_basic,prob_delta)
% USFFT_ALGORITHM uniform sFFT based on random rank-1 lattice sampling.
% This is an approximation algorithm for the reconstruction of sparse
% high-dimensional trigonometric polynomials, which adaptively constructs the
% index set of frequencies belonging to the non-zero Fourier coefficients in a
% dimension incremental way. For further details, please read
%
%   L. Kaemmerer, D. Potts and F. Taubert
%   The uniform sparse FFT with application to PDEs with random coefficients
%   Submitted, 2021
%
%   Input parameters:
%      fct_handle - function handle (black box) which accepts sampling
%                   nodes as an array of size [num_nodes,d]
%      d - dimensionality
%      N - parameter for search space Gamma = [-N,...,N]^d of possible
%          frequencies
%      algorithm_name - (optional) type of algorithm,
%                       either 'rlfftriabs' (default) for Algorithm 1
%                       or 'rlfftl4abs' for Algorithm 2
%      sparsity_s - (optional) target sparsity of output, default (2*N+1)^d
%      sparsity_s_local - (optional) value of sparsity_s in local loops,
%                         uses the value of sparsity_s as default
%      threshold_theta - (optional) relative threshold theta used for
%                        truncation in step 2e, default is 1e-12
%      threshold_theta_basic - (optional) relative threshold theta used for
%                              truncation in steps 1 and 2a,
%                              uses the value of threshold_theta as default 
%      niter_r - (optional) number of detection iteration r in steps 2b-2e,
%                default is 1
%      niter_r_basic - (optional) number of detection iteration r in
%                      steps 1 and 2a, uses the value of niter_r as default
%      prob_delta - (optional) probability parameter for choosing L,
%                   default is 0.1
%
%   Output parameters:
%      nodes - matrix of the spatial nodes in [0,1]^2, size [d_spatial,#nodes]
%      freq - detected frequencies belonging to the (approximately)
%             sparsity_s largest Fourier coefficients, size [#freq,d]
%      val - corresponding Fourier coefficients for each spatial node
%      freq_cand_sizes - array of length d with number of investigated
%                        frequencies in each dimension increment step t
%      sample_sizes - array of length d with number of sampling locations
%                     for the function fct_handle in each dimension 
%                     increment step t
%      run_times - array of length 5 with runtimes measured for various
%                  parts of the algorithm,
%                  run_times(1): steps 1 and 2a,
%                  run_times(2): building of reconstructing rank-1 lattices
%                  run_times(3): sampling step 2c
%                  run_times(4): inverse rank-1 lattice FFT step 2d
%                  run_times(5): total runtime of the algorithm
%
% Copyright (c) 2021 Fabian Taubert, Lutz Kaemmerer
%
% This program is free software; you can redistribute it and/or modify it under
% the terms of the GNU General Public License as published by the Free Software
% Foundation; either version 2 of the License, or (at your option) any later
% version.
%
% This program is distributed in the hope that it will be useful, but WITHOUT
% ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
% details.
%
% You should have received a copy of the GNU General Public License along with
% this program; if not, write to the Free Software Foundation, Inc., 51
% Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

if nargin < 4 || isempty(algorithm_name)
  algorithm_name = 'rlfftriabs';
end

if nargin < 5 || isempty(sparsity_s)
  sparsity_s = (2*N+1)^d;
end

if nargin < 6 || isempty(sparsity_s_local)
  sparsity_s_local = sparsity_s;
end

if nargin < 7 || isempty(threshold_theta)
  threshold_theta = 1e-12;
end

if nargin < 8 || isempty(threshold_theta_basic)
  threshold_theta_basic = threshold_theta;
end

if nargin < 9 || isempty(niter_r)
  niter_r = 1;
end

if nargin < 10 || isempty(niter_r_basic)
  niter_r_basic = niter_r;
end

if (nargin < 11) || isempty(prob_delta)
  prob_delta = 0.1;
end

check_naturalnumber_or_error(d, 'dimensionality d');
check_naturalnumber_or_error(N, 'parameter N');

switch algorithm_name
  case {'rlfftriabs','rlfftl4abs'}
  otherwise
    error('unsupported algorithm: %s', algorithm_name);
end

check_naturalnumber_or_error(sparsity_s, 'parameter sparsity_s');
check_naturalnumber_or_error(sparsity_s_local, 'parameter sparsity_s_local');
check_0leq_param_leq1_or_error(threshold_theta, 'parameter threshold_theta');
check_0leq_param_leq1_or_error(threshold_theta_basic, 'parameter threshold_theta_basic');
check_naturalnumber_or_error(niter_r, 'parameter niter_r');
check_naturalnumber_or_error(niter_r_basic, 'parameter niter_r_basic');

fprintf('sfft: d = %d, N = %d, algorithm = %s, sparsity_s = %d, sparsity_s_local = %d\n', d, N, algorithm_name, sparsity_s, sparsity_s_local);
fprintf('sfft: threshold_theta = %.3e, basic_threshold_theta = %.3e, niter_r = %d, niter_r_basic = %d\n', threshold_theta, threshold_theta_basic, niter_r, niter_r_basic);

freq_cand_sizes = zeros(1,d);
sample_sizes = zeros(1,d);
run_times = zeros(1,5); % basic, lattice search, sampling, lfft, all

freq_cand_sizes(1) = 2*N+1;
num_R1L_sizes = zeros(1,d);

tAll = tic;

if d == 1
  error('d=1 is not supported in this implementation');
else
  freq = zeros(1,1);
  
  % step 1
  L_1 = 2*N+1;
  for vv=1:niter_r_basic
    tBasic = tic;
    [freq_local,grid_size] = reco_cardI_basic1_delta(fct_handle, d, N, 1, threshold_theta_basic);
    run_times(1) = run_times(1) + toc(tBasic);
    sample_sizes(1) = sample_sizes(1) + 2*N+1;
    freq = union(freq,freq_local,'rows');
  end 
  
  fprintf('step 1: L_1 = %d, card(I^(1)) = %d\n', L_1, size(freq,1));
      
  % step 2
  for t = 2:d
    fprintf('step 2: t = %d start\n', t);
    freq_t = zeros(1,1);
    % step 2a
    L_t = 2*N+1;
    for vv=1:niter_r_basic
      tBasic = tic;
      freqnew_local = reco_cardI_basic1_delta(fct_handle, d, N, t, threshold_theta_basic);
      run_times(1) = run_times(1) + toc(tBasic);
      sample_sizes(t) = sample_sizes(t) + 2*N+1;
      freq_t = union(freq_t,freqnew_local,'rows');
    end
    fprintf('step 2a: L_%d = %d, card(I^(%d)) = %d\n', t, L_t, t, size(freq_t,1));
    A=reshape(repmat(1:size(freq,1), size(freq_t,1), 1), size(freq,1)*size(freq_t,1), 1);
    B=repmat((1:size(freq_t,1))', size(freq,1), 1);
    indcross = [A B];
    clear A B;
    freq_cand = [freq(indcross(:,1),:) freq_t(indcross(:,2),:)];
    clear indcross;

    freq_cand_sizes(t) = size(freq_cand,1);

    % step 2b
    if t==d
      niter_r_tilde=1;
    else
      niter_r_tilde=niter_r;
    end

    freq = [];
    val = [];

    tLattice = tic;
    c = 7;
    lambda = max(2*N+1, floor(c*sparsity_s));
    M = simplenextprime(lambda);
    if prob_delta < 0
      num_R1L = ceil(abs(prob_delta) * c*(c-2)/(c*0.5-1)^2/log(c-1) * (log(abs(size(freq_cand,1)))));
    else
      num_R1L = ceil(c*(c-2)/(c*0.5-1)^2/log(c-1) * (log(abs(size(freq_cand,1)))-real(log(prob_delta))));
    end
    if num_R1L < 3
      num_R1L = 3;
    end
    if mod(num_R1L,2) == 0
      num_R1L = num_R1L+1;
    end
    Ms = repmat(M, 1, num_R1L);
    zs = randi(M, num_R1L, t) - 1;
    run_times(2) = run_times(2) + toc(tLattice);
    fprintf('step 2b: finished building reconstructing rank-1 lattice\n');
    fprintf('step 2b: %d lattices, %d samples\n', length(Ms), 1+sum(Ms)-length(Ms));
    fprintf('step 2c: %d samples, r_tilde = %d detection iterations\n', 1+sum(Ms)-length(Ms), niter_r_tilde);
    num_R1L_sizes(t) = num_R1L;

    ind = [];
    for vv=1:niter_r_tilde
      samples = cell(grid_size-1,num_R1L);
      rnd_vals = rand(d,1);

      tSampling = tic;
      for ilattice = 1:length(Ms)
        z = zs(ilattice,:);
        M = Ms(ilattice);
        cur_rnd_vals = rnd_vals(:,1);
        xj = repmat(cur_rnd_vals.',M,1);
        xj(:,1:t) = mod((0:M-1)' * z(1:t), M)/M;

        [nodes,samples_nodal_full] = fct_handle(xj);
        inner_ind = find(all([nodes(:,:) == 1 ; nodes(:,:) == 0]==0)); %exclude boundary nodes
        inner_nodes = nodes(:,inner_ind);
        samples_grid_temp = samples_nodal_full(:,inner_ind);
        grid_size = size(inner_nodes,2);
        clear xj;

        for g=1:grid_size  
            samples_grid{g,ilattice} = samples_grid_temp(:,g); 
        end
        sample_sizes(t) = sample_sizes(t) + size(samples_grid_temp,1);
        clear samples_grid_temp z M;
      end
      run_times(3) = run_times(3) + toc(tSampling);

      freq_grid = zeros(0,t);
      for g = 1:grid_size
          samples = samples_grid(g,:);

          % step 2d
          tLFFT = tic;
          switch algorithm_name
            case {'rlfftriabs'}
              [freq_local, val_local] = sfft_random_r1l_alg2ri(freq_cand, zs, Ms, samples, threshold_theta);
            case {'rlfftl4abs'}
              [freq_local, val_local] = sfft_random_r1l_alg2l4(freq_cand, zs, Ms, samples, threshold_theta);
          end
          run_times(4) = run_times(4) + toc(tLFFT);
          clear samples;

          % step 2e  
          ind = find(abs(val_local) >= threshold_theta);
          freq_local = freq_local(ind,:);
          val_local = val_local(ind);
          if size(freq_local,1) > sparsity_s_local
            [~,ind_local] = sort(abs(val_local),'descend');
            freq_local = freq_local(ind_local(1:sparsity_s_local),:);
            clear ind_local;
          end
          clear val_local;

          freq_grid = union(freq_grid,freq_local,'rows');
      end
      if vv==1
        freq = freq_grid;
      else
        freq = union(freq,freq_grid,'rows');
      end
      clear freq_local freq_grid samples_grid; 

      fprintf('step 2e: i = %d, card(I^(1,...,%d)) = %d\n', vv, t, size(freq,1));
    end
    fprintf('step 2e: result dim %d of %d: %d freq candidates, %d samples, %d freq detected\n', t, d, size(freq_cand,1), 1+sum(Ms)-length(Ms), size(freq,1));
    clear ind;
    fprintf('step 2: t = %d end, d = %d, nsamples = %d\n\n', t, d, 1+sum(Ms)-length(Ms));
  end

end

%step 3
optcmr1l.c = 2;
optcmr1l.delta = 0.5;
[Ms,zs] = multi_lattice_construct_mr1l(freq,2,optcmr1l);
samples_grid = [];

tSampling = tic;
for ilattice = 1:length(Ms)
  z = zs(ilattice,:);
  M = Ms(ilattice);
  cur_rnd_vals = rnd_vals(:,1);
  xj = repmat(cur_rnd_vals.',M,1);
  xj(:,1:t) = mod((0:M-1)' * z(1:t), M)/M;

  [nodes,samples_nodal_full] = fct_handle(xj);
  samples_grid_temp = samples_nodal_full;
  full_grid_size = size(nodes,2);
  clear xj;

  if ilattice == 1
    samples_grid = [samples_grid; samples_grid_temp];
  else
    samples_grid = [samples_grid; samples_grid_temp(2:end,:)];
  end
  clear samples_grid_temp z M;
end
run_times(3) = run_times(3) + toc(tSampling);
sample_sizes(t) = sample_sizes(t) + size(samples_grid,1);

val = zeros(size(freq,1),full_grid_size);

for g = 1:full_grid_size
    tLFFT = tic;
    samples = samples_grid(:,g);
    val(:,g) = multi_lattice_ifft_noniter_avg(samples,freq,Ms,zs);
    run_times(4) = run_times(4) + toc(tLFFT);
end

fprintf('step 3: %.10g possible freq., %d total samples, %d freq. detected\n\n', (2*N+1)^d, sum(sample_sizes), size(freq,1));
run_times(end) = toc(tAll);

end

function [val] = check_naturalnumber_or_error(param, name)
  val = 0;
  if (param < 1) || (param ~= int32(param))
    error('%s must be a natural number', name);
  end
end

function [val] = check_0leq_param_leq1_or_error(param, name)
  val = 0;
  if (param < 0) || (param > 1)
    error('%s must be >= 0 and <= 1', name);
  end
end

function [filename] = check_mex_file_or_warn(filename_without_ext)
  filename = strcat(filename_without_ext,strcat('.',mexext));
  if exist(filename, 'file') ~= 3
    warning('MATLAB mex file %s is missing. The algorithm will be very slow!\nPlease use the MATLAB script compile_mex_files_without_openmp.m or compile_mex_files_with_openmp.m', filename);
  end
end